#!/usr/bin/env python3
#
# Copyright 2022 Flant JSC Licensed under Apache License 2.0
#

import json
import os
import sys
from dataclasses import dataclass
from typing import Iterable

from dictdiffer import deepcopy
from dotmap import DotMap

from .kubernetes import KubeOperationCollector
from .metrics import MetricsCollector
from .storage import FileStorage
from .values import ValuesPatchesCollector


class Output:
    """
    Container for output means for metrics, kubernetes, and values.

    Metrics, Kubernetes JSON patches and values JSON patches are collected in underlying storages,
    whether shell-operator (or addon-operator) file paths, or into memory.
    """

    # Values with outputs for tests, values patches are less convenient than values
    # themselves.
    values: DotMap = None

    def __init__(self, metrics, kube_operations, values_patches):
        self.metrics = metrics
        self.kube_operations = kube_operations
        self.values_patches = values_patches

    # TODO  logger: --log-proxy-hook-json / LOG_PROXY_HOOK_JSON (default=false)
    #
    # Delegate hook stdout/stderr JSON logging to the hooks and act as a proxy that adds some extra
    # fields before just printing the output. NOTE: It ignores LOG_TYPE for the output of the hooks;
    # expects JSON lines to stdout/stderr from the hooks

    def flush(self):
        file_outputs = (
            ("METRICS_PATH", self.metrics),
            ("KUBERNETES_PATCH_PATH", self.kube_operations),
            ("VALUES_JSON_PATCH_PATH", self.values_patches),
        )

        for path_env, collector in file_outputs:
            path = os.getenv(path_env)
            if not path:
                # No values in Shell Operator
                continue
            with FileStorage(path) as file:
                for payload in collector.data:
                    file.write(payload)


@dataclass
class Context:
    def __init__(self, binding_context: dict, initial_values: dict, output: Output):
        self.binding_context = binding_context
        self.snapshots = binding_context.get("snapshots", {})
        self.values = DotMap(deepcopy(initial_values))  # DotMap for values.dot.notation
        self.output = output

    @property
    def metrics(self):
        return self.output.metrics

    @property
    def kubernetes(self):
        return self.output.kube_operations

    @property
    def values_patches(self):
        return self.output.values_patches


def read_binding_context_file():
    """
    Iterates over hook contexts in the binding context file.

    :yield ctx: hook binding context
    """
    context_path = os.getenv("BINDING_CONTEXT_PATH")
    with open(context_path, "r", encoding="utf-8") as f:
        context = json.load(f)
    for ctx in context:
        yield ctx


def read_values_file():
    """
    Reads module values from the values file.

    :return values: the dict of the values
    """
    values_path = os.getenv("VALUES_PATH")
    if not values_path:
        # No values in Shell Operator
        return {}
    with open(values_path, "r", encoding="utf-8") as f:
        values = json.load(f)
    return values


def __run(func, binding_context: list, initial_values: dict):
    """
    Run the hook function with config. Accepts config path or config text.

    :param func: the function to run
    :param binding_context: the list of hook binding contexts
    :param initial_values: initial values
    :return output: output means with all generated payloads and updated values
    """

    if not binding_context:
        binding_context = [{}]
    if not initial_values:
        initial_values = DotMap()

    output = Output(
        MetricsCollector(),
        KubeOperationCollector(),
        ValuesPatchesCollector(initial_values),
    )

    for bindctx in binding_context:
        hookctx = Context(bindctx, initial_values, output)
        func(hookctx)
        output.values = hookctx.values
        output.values_patches.update(hookctx.values)

    return output


def run(func, configpath=None, config=None):
    """
    Run the hook function with config. Accepts config path or config text.

    :param configpath: path to the hook config file
    :param config: hook config text itself
    """

    if len(sys.argv) > 1 and sys.argv[1] == "--config":
        if config is None and configpath is None:
            raise ValueError("config or configpath must be provided")

        if config is not None:
            print(config)
        else:
            with open(configpath, "r", encoding="utf-8") as cf:
                print(cf.read())

        sys.exit(0)

    binding_context = read_binding_context_file()
    initial_values = read_values_file()

    output = __run(func, binding_context, initial_values)

    output.flush()


def testrun(
    func,
    binding_context: Iterable = None,
    initial_values: dict = None,
) -> Output:
    """
    Test-run the hook function. Accepts binding context and initial values.

    Returns output means for metrics, kubernetes, values patches, and also modified values for more
    convenient tests.

    :param binding_context: the list of hook binding contexts
    :param initial_values: initial values
    :return: output means for metrics and kubernetes
    """

    output = __run(func, binding_context, initial_values)
    return output



#!/usr/bin/env python3
#
# Copyright 2022 Flant JSC Licensed under Apache License 2.0
#


class KubeOperationCollector:
    """
    Wrapper for the kubernetes actions: creation, deletion, patching.
    """

    def __init__(self):
        self.data = []

    def collect(self, payload: dict):
        self.data.append(payload)

    def create(self, obj):
        """
        :param obj: must be serializable to JSON
        """
        self.collect({"operation": "Create", "object": obj})

    def create_or_update(self, obj):
        """
        :param obj: must be serializable to JSON
        """
        self.collect({"operation": "CreateOrUpdate", "object": obj})

    def create_if_not_exists(self, obj):
        """
        :param obj: must be serializable to JSON
        """
        self.collect({"operation": "CreateIfNotExists", "object": obj})

    def __delete(
        self, operation, kind, namespace, name, apiVersion=None, subresource=None
    ):
        """
        :param kind: object Kind.
        :param namespace: object namespace. If empty, implies operation on a cluster-level resource.
        :param name: object name.
        :param apiVersion: optional field that specifies object apiVersion. If not present, we'll
            use preferred apiVersion for the given kind.
        :param subresource: a subresource name if subresource is to be transformed. For example,
            status.
        """
        if operation not in ("Delete", "DeleteInBackground", "DeleteNonCascading"):
            raise ValueError(f"Invalid delete operation: {operation}")

        obj = {
            "kind": kind,
            "namespace": namespace,
            "name": name,
        }
        if apiVersion is not None:
            obj["apiVersion"] = apiVersion
        if subresource is not None:
            obj["subresource"] = subresource

        self.collect({"operation": operation, "object": obj})

    def delete(self, kind, namespace, name, apiVersion=None, subresource=None):
        """
        :param kind: object Kind.
        :param namespace: object namespace. If empty, implies operation on a cluster-level resource.
        :param name: object name.
        :param apiVersion: optional field that specifies object apiVersion. If not present, we'll
            use preferred apiVersion for the given kind.
        :param subresource: a subresource name if subresource is to be transformed. For example,
            status.
        """
        return self.__delete("Delete", kind, namespace, name, apiVersion, subresource)

    def delete_in_backgroud(
        self, kind, namespace, name, apiVersion=None, subresource=None
    ):
        """
        :param kind: object Kind.
        :param namespace: object namespace. If empty, implies operation on a cluster-level resource.
        :param name: object name.
        :param apiVersion: optional field that specifies object apiVersion. If not present, we'll
            use preferred apiVersion for the given kind.
        :param subresource: a subresource name if subresource is to be transformed. For example,
            status.
        """
        return self.__delete(
            "DeleteInBackground", kind, namespace, name, apiVersion, subresource
        )

    def delete_non_cascading(
        self, kind, namespace, name, apiVersion=None, subresource=None
    ):
        """
        :param kind: object Kind.
        :param namespace: object namespace. If empty, implies operation on a cluster-level resource.
        :param name: object name.
        :param apiVersion: optional field that specifies object apiVersion. If not present, we'll
            use preferred apiVersion for the given kind.
        :param subresource: a subresource name if subresource is to be transformed. For example,
            status.
        """
        return self.__delete(
            "DeleteNonCascading", kind, namespace, name, apiVersion, subresource
        )

    def merge_patch(
        self,
        kind,
        namespace,
        name,
        patch: dict,
        apiVersion=None,
        subresource=None,
        ignoreMissingObject=False,
    ):
        """
        :param operation: specifies an operation's type.
        :param apiVersion: optional field that specifies object apiVersion. If not present, we'll
            use preferred apiVersion for the given kind.
        :param kind: object Kind.
        :param namespace: object Namespace. If empty, implies operation on a Cluster-level
            resource.
        :param name: object name.
        :param patch: describes transformations to perform on an object. Can be a normal JSON or
            YAML array or a stringified JSON or YAML array.
        :param subresource: a subresource name if subresource is to be transformed. For example,
            status.
        :param ignoreMissingObject: set to true to ignore error when patching non existent object.
        """
        return self.__patch(
            "MergePatch",
            kind,
            namespace,
            name,
            patch,
            apiVersion,
            subresource,
            ignoreMissingObject,
        )

    def json_patch(
        self,
        kind,
        namespace,
        name,
        patch: dict,
        apiVersion=None,
        subresource=None,
        ignoreMissingObject=False,
    ):
        """
        :param apiVersion: optional field that specifies object apiVersion. If not present, we'll
            use preferred apiVersion for the given kind.
        :param kind: object Kind.
        :param namespace: object Namespace. If empty, implies operation on a Cluster-level resource.
        :param name: object name.
        :param patch: describes transformations to perform on an object. Can be a normal JSON or
            YAML array or a stringified JSON or YAML array.
        :param subresource: a subresource name if subresource is to be transformed. For example,
            status.
        :param ignoreMissingObject: set to true to ignore error when patching non existent object.
        """
        return self.__patch(
            "JSONPatch",
            kind,
            namespace,
            name,
            patch,
            apiVersion,
            subresource,
            ignoreMissingObject,
        )

    def __patch(
        self,
        operation,
        kind,
        namespace,
        name,
        patch: dict,
        apiVersion=None,
        subresource=None,
        ignoreMissingObject=False,
    ):
        """
        :param operation: specifies an operation's type.
        :param apiVersion: optional field that specifies object apiVersion. If not present, we'll
            use preferred apiVersion for the given kind.
        :param kind: object Kind.
        :param namespace: object Namespace. If empty, implies operation on a Cluster-level resource.
        :param name: object name.
        :param patch: describes transformations to perform on an object. Can be a normal JSON or
            YAML array or a stringified JSON or YAML array.
        :param subresource: a subresource name if subresource is to be transformed. For example,
            status.
        :param ignoreMissingObject: set to true to ignore error when patching non existent object.
        """
        ret = {
            "operation": operation,
            "kind": kind,
            "namespace": namespace,
            "name": name,
        }
        if operation == "MergePatch":
            ret["mergePatch"] = patch
        elif operation == "JSONPatch":
            ret["jsonPatch"] = patch
        else:
            raise ValueError(f"Invalid patch operation: {operation}")

        if apiVersion is not None:
            ret["apiVersion"] = apiVersion
        if subresource is not None:
            ret["subresource"] = subresource
        if ignoreMissingObject:
            ret["ignoreMissingObject"] = ignoreMissingObject

        self.collect(ret)



#!/usr/bin/env python3
#
# Copyright 2022 Flant JSC Licensed under Apache License 2.0
#

import json


class FileStorage:
    """
    Context manager wrapping the appending JSON per line to file
    """

    def __init__(self, path):
        self.file = open(path, "a", encoding="utf-8")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()

    def write(self, payload: dict):
        self.file.write(json.dumps(payload))
        self.file.write("\n")



#!/usr/bin/env python3
#
# Copyright 2023 Flant JSC Licensed under Apache License 2.0
#

from typing import Iterable

from dictdiffer import deepcopy, diff


class ValuesPatchesCollector:
    """
    Wrapper for the values manipulations (JSON patches)
    """

    def __init__(self, values: dict):
        self.initial_values = deepcopy(values)
        self.data = []

    def collect(self, payload: dict):
        self.data.append(payload)

    def update(self, updated_values: dict):
        for patch in values_json_patches(self.initial_values, updated_values):
            self.collect(patch)


def values_json_patches(initial_values: dict, updated_values: dict):
    changes = diff(
        initial_values,
        updated_values,
        dot_notation=False,  # always return path as list
        expand=True,  # do not compact values in single operation
    )
    for change in changes:
        yield json_patch(change)


def json_patch(change):
    """
    Trtansform dictdiffer change to JSON patch.
    https://jsonpatch.com/#operations
    """
    op, path_segments, values = change

    if op == "add":
        #   op    |______path_________|   value
        #    |    |                   |  /
        # ('add', ['x', 'y', 'a'], [(2, 2)])

        key, value = values[0]
        path = json_path(path_segments + [key])
        return {"op": "add", "path": path, "value": value}

    if op == "change":
        #   op       |______path______|  from  to
        #    |       |                |   |   /
        # ('change', ['x', 'y', 'a', 0], (1, 0))

        value = values[1]
        path = json_path(path_segments)
        return {"op": "replace", "path": path, "value": value}

    if op == "remove":
        #   op       |______path_____|     value
        #    |       |               |    /
        # ('remove', ['x', 'y'], [('t', 0)])

        key = values[0][0]
        path = json_path(path_segments + [key])
        return {"op": "remove", "path": path}

    raise ValueError(f"Unknown patch operation: {op}")


def json_path(path: Iterable):
    return "/" + "/".join([str(p) for p in path])



#!/usr/bin/env python3
#
# Copyright 2022 Flant JSC Licensed under Apache License 2.0
#


class MetricsCollector:
    """
    Wrapper for metrics exporting. Accepts raw dicts and appends them into the metrics file.
    """

    def __init__(self):
        self.data = []

    def collect(self, payload: dict):
        self.data.append(payload)

    def expire(self, group: str):
        """Expire all metrics in the group.

        Args:
            group (str): metric group name
        """
        self.collect({"action": "expire", "group": group})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['shell_operator']

package_data = \
{'': ['*']}

install_requires = \
['dictdiffer>=0.9.0,<0.10.0', 'dotmap>=1.3.30,<2.0.0']

setup_kwargs = {
    'name': 'shell-operator',
    'version': '0.0.9',
    'description': 'Python SDK for Shell Operator hooks',
    'long_description': '# Shell Operator Python Framework\n\nSimplifies writing [shell-operator](https://github.com/flant/shell-operator) hooks in Python.\n\n## Install\n\n```bash\npip install shell-operator\n```\n\n## Sample hook\n\n```python\n# hello.py\nfrom shell_operator import hook\n\ndef main(ctx: hook.Context):\n    # Manipulate kubernetes state\n    # ... object = { "kind" : "Pod", "apiVersion" : "v1", ... }\n    ctx.kubernetes.create_or_update(object)\n\n    # Export metrics\n    # ... metric = { "name" : "power", "group": "my_hook", "set" : 9000, ... }\n    ctx.metrics.collect(metric)\n\n    # Use in-memory values for helm chart. Shell Operator does not support values, but Addon Operator and Deckhouse do.\n    ctx.values.myModule.deployment.replicas = 5\n\n\nif __name__ == "__main__":\n    hook.run(main, configpath="hello.yaml") # \'config\' arg is also supported for raw string\n```\n\n```yaml\n# hello.yaml\nconfigVersion: v1\nonStartup: 10\n```\n\n## How to test\n\nAn example for pytest\n\n```python\n# hello_test.py\n\nfrom hello import main\nfrom shell_operator import hook\n\n# Inputs\n#   initial_values = { ... }\n#   binding_context = [ { ... } ]\n# Outputs\n#   expected_metrics = [ ... ]\n#   expected_kube_operations = [ ... ]\n#   expected_values_patches = [ ... ]\n#   expected_values = { ... }\n\ndef test_hello():\n    out = hook.testrun(main, binding_context, initial_values)\n\n    assert out.metrics.data == expected_metrics\n    assert out.kube_operations.data == expected_kube_operations\n    assert out.values_patches.data == expected_values_patches\n\n    assert out.values.myModule.deployment.repicas == 5\n```\n',
    'author': 'Eugene Shevchenko',
    'author_email': 'evgeny.shevchenko@flant.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)
